@@ -1,4 +1,4 @@
-/* Copyright (c) 2016, 2020, MariaDB Corporation.
+/* Copyright (c) 2016, 2021, MariaDB Corporation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -18,7 +18,22 @@
 #include "sql_priv.h"
 #include "sql_class.h"
 #include "item.h"
+#include "sql_parse.h" // For check_stack_overrun
 
+/*
+  Allocating memory and *also* using it (reading and
+  writing from it) because some build instructions cause
+  compiler to optimize out stack_used_up. Since alloca()
+  here depends on stack_used_up, it doesnt get executed
+  correctly and causes json_debug_nonembedded to fail
+  ( --error ER_STACK_OVERRUN_NEED_MORE does not occur).
+*/
+#define ALLOCATE_MEM_ON_STACK(A) do \
+                              { \
+                                uchar *array= (uchar*)alloca(A); \
+                                bzero(array, A); \
+                                my_checksum(0, array, A); \
+                              } while(0)
 
 /*
   Compare ASCII string against the string with the specified
@@ -118,7 +133,7 @@ static const char tab_arr[TAB_SIZE_LIMIT+1]= "        ";
 
 static int append_tab(String *js, int depth, int tab_size)
 {
-  if (js->append("\n", 1))
+  if (js->append('\n'))
     return 1;
   for (int i=0; i<depth; i++)
   {
@@ -128,21 +143,131 @@ static int append_tab(String *js, int depth, int tab_size)
   return 0;
 }
 
+int json_path_parts_compare(
+    const json_path_step_t *a, const json_path_step_t *a_end,
+    const json_path_step_t *b, const json_path_step_t *b_end,
+    enum json_value_types vt)
+{
+  int res, res2;
+
+  DBUG_EXECUTE_IF("json_check_min_stack_requirement",
+                  {
+                    long arbitrary_var;
+                    long stack_used_up= (available_stack_size(current_thd->thread_stack, &arbitrary_var));
+                    ALLOCATE_MEM_ON_STACK(my_thread_stack_size-stack_used_up-STACK_MIN_SIZE);
+                  });
+  if (check_stack_overrun(current_thd, STACK_MIN_SIZE , NULL))
+    return 1;
+
+  while (a <= a_end)
+  {
+    if (b > b_end)
+    {
+      while (vt != JSON_VALUE_ARRAY &&
+             (a->type & JSON_PATH_ARRAY_WILD) == JSON_PATH_ARRAY &&
+             a->n_item == 0)
+      {
+        if (++a > a_end)
+          return 0;
+      }
+      return -2;
+    }
+
+    DBUG_ASSERT((b->type & (JSON_PATH_WILD | JSON_PATH_DOUBLE_WILD)) == 0);
+
+
+    if (a->type & JSON_PATH_ARRAY)
+    {
+      if (b->type & JSON_PATH_ARRAY)
+      {
+        if ((a->type & JSON_PATH_WILD) || a->n_item == b->n_item)
+          goto step_fits;
+        goto step_failed;
+      }
+      if ((a->type & JSON_PATH_WILD) == 0 && a->n_item == 0)
+        goto step_fits_autowrap;
+      goto step_failed;
+    }
+    else /* JSON_PATH_KEY */
+    {
+      if (!(b->type & JSON_PATH_KEY))
+        goto step_failed;
+
+      if (!(a->type & JSON_PATH_WILD) &&
+          (a->key_end - a->key != b->key_end - b->key ||
+           memcmp(a->key, b->key, a->key_end - a->key) != 0))
+        goto step_failed;
+
+      goto step_fits;
+    }
+step_failed:
+    if (!(a->type & JSON_PATH_DOUBLE_WILD))
+      return -1;
+    b++;
+    continue;
+
+step_fits:
+    b++;
+    if (!(a->type & JSON_PATH_DOUBLE_WILD))
+    {
+      a++;
+      continue;
+    }
+
+    /* Double wild handling needs recursions. */
+    res= json_path_parts_compare(a+1, a_end, b, b_end, vt);
+    if (res == 0)
+      return 0;
+
+    res2= json_path_parts_compare(a, a_end, b, b_end, vt);
+
+    return (res2 >= 0) ? res2 : res;
+
+step_fits_autowrap:
+    if (!(a->type & JSON_PATH_DOUBLE_WILD))
+    {
+      a++;
+      continue;
+    }
+
+    /* Double wild handling needs recursions. */
+    res= json_path_parts_compare(a+1, a_end, b+1, b_end, vt);
+    if (res == 0)
+      return 0;
+
+    res2= json_path_parts_compare(a, a_end, b+1, b_end, vt);
+
+    return (res2 >= 0) ? res2 : res;
+
+  }
+
+  return b <= b_end;
+}
+
+
+int json_path_compare(const json_path_t *a, const json_path_t *b,
+                      enum json_value_types vt)
+{
+  return json_path_parts_compare(a->steps+1, a->last_step,
+                                 b->steps+1, b->last_step, vt);
+}
+
 
 static int json_nice(json_engine_t *je, String *nice_js,
                      Item_func_json_format::formats mode, int tab_size=4)
 {
   int depth= 0;
-  const char *comma, *colon;
+  static const char *comma= ", ", *colon= "\": ";
   uint comma_len, colon_len;
   int first_value= 1;
 
-  DBUG_ASSERT(je->s.cs == nice_js->charset());
+  nice_js->length(0);
+  nice_js->set_charset(je->s.cs);
+  nice_js->alloc(je->s.str_end - je->s.c_str + 32);
+
   DBUG_ASSERT(mode != Item_func_json_format::DETAILED ||
               (tab_size >= 0 && tab_size <= TAB_SIZE_LIMIT));
 
-  comma= ", ";
-  colon= "\": ";
   if (mode == Item_func_json_format::LOOSE)
   {
     comma_len= 2;
@@ -183,7 +308,7 @@ static int json_nice(json_engine_t *je, String *nice_js,
             append_tab(nice_js, depth, tab_size))
           goto error;
 
-        nice_js->append("\"", 1);
+        nice_js->append('"');
         append_simple(nice_js, key_start, key_end - key_start);
         nice_js->append(colon, colon_len);
       }
@@ -239,7 +364,7 @@ handle_value:
     };
   } while (json_scan_next(je) == 0);
 
-  return je->s.error;
+  return je->s.error || *je->killed_ptr;
 
 error:
   return 1;
@@ -397,7 +522,7 @@ bool Item_func_json_exists::fix_length_and_dec()
 {
   if (Item_bool_func::fix_length_and_dec())
     return TRUE;
-  maybe_null= 1;
+  set_maybe_null();
   path.set_constant_flag(args[1]->const_item());
   return FALSE;
 }
@@ -451,7 +576,7 @@ bool Item_func_json_value::fix_length_and_dec()
   collation.set(args[0]->collation);
   max_length= args[0]->max_length;
   set_constant_flag(args[1]->const_item());
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -461,7 +586,7 @@ bool Item_func_json_query::fix_length_and_dec()
   collation.set(args[0]->collation);
   max_length= args[0]->max_length;
   set_constant_flag(args[1]->const_item());
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -591,9 +716,9 @@ String *Item_func_json_quote::val_str(String *str)
   str->length(0);
   str->set_charset(&my_charset_utf8mb4_bin);
 
-  if (str->append("\"", 1) ||
+  if (str->append('"') ||
       st_append_escaped(str, s) ||
-      str->append("\"", 1))
+      str->append('"'))
   {
     /* Report an error. */
     null_value= 1;
@@ -609,7 +734,7 @@ bool Item_func_json_unquote::fix_length_and_dec()
   collation.set(&my_charset_utf8mb3_general_ci,
                 DERIVATION_COERCIBLE, MY_REPERTOIRE_ASCII);
   max_length= args[0]->max_length;
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -730,7 +855,7 @@ bool Item_func_json_extract::fix_length_and_dec()
   max_length= args[0]->max_length * (arg_count - 1);
 
   mark_constant_paths(paths, args+1, arg_count-1);
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -805,7 +930,7 @@ String *Item_func_json_extract::read_json(String *str,
     str->set_charset(js->charset());
     str->length(0);
 
-    if (possible_multiple_values && str->append("[", 1))
+    if (possible_multiple_values && str->append('['))
       goto error;
   }
 
@@ -867,14 +992,12 @@ String *Item_func_json_extract::read_json(String *str,
     goto return_null;
   }
 
-  if (possible_multiple_values && str->append("]", 1))
+  if (possible_multiple_values && str->append(']'))
     goto error; /* Out of memory. */
 
   js= str;
   json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                   (const uchar *) js->ptr() + js->length());
-  tmp_js.length(0);
-  tmp_js.set_charset(js->charset());
   if (json_nice(&je, &tmp_js, Item_func_json_format::LOOSE))
     goto error;
 
@@ -985,13 +1108,15 @@ my_decimal *Item_func_json_extract::val_decimal(my_decimal *to)
       case JSON_VALUE_OBJECT:
       case JSON_VALUE_ARRAY:
       case JSON_VALUE_FALSE:
-      case JSON_VALUE_NULL:
       case JSON_VALUE_UNINITALIZED:
-      break;
+      // TODO: fix: NULL should be NULL
+      case JSON_VALUE_NULL:
+        int2my_decimal(E_DEC_FATAL_ERROR, 0, false/*unsigned_flag*/, to);
+        return to;
     };
   }
-  int2my_decimal(E_DEC_FATAL_ERROR, 0, false/*unsigned_flag*/, to);
-  return to;
+  DBUG_ASSERT(null_value);
+  return 0;
 }
 
 
@@ -1000,7 +1125,7 @@ bool Item_func_json_contains::fix_length_and_dec()
 {
   a2_constant= args[1]->const_item();
   a2_parsed= FALSE;
-  maybe_null= 1;
+  set_maybe_null();
   if (arg_count > 2)
     path.set_constant_flag(args[2]->const_item());
   return Item_bool_func::fix_length_and_dec();
@@ -1029,6 +1154,14 @@ static int check_contains(json_engine_t *js, json_engine_t *value)
 {
   json_engine_t loc_js;
   bool set_js;
+  DBUG_EXECUTE_IF("json_check_min_stack_requirement",
+                  {
+                    long arbitrary_var;
+                    long stack_used_up= (available_stack_size(current_thd->thread_stack, &arbitrary_var));
+                    ALLOCATE_MEM_ON_STACK(my_thread_stack_size-stack_used_up-STACK_MIN_SIZE);
+                  });
+  if (check_stack_overrun(current_thd, STACK_MIN_SIZE , NULL))
+    return 1;
 
   switch (js->value_type)
   {
@@ -1251,7 +1384,7 @@ bool Item_func_json_contains_path::fix_length_and_dec()
 {
   ooa_constant= args[1]->const_item();
   ooa_parsed= FALSE;
-  maybe_null= 1;
+  set_maybe_null();
   mark_constant_paths(paths, args+2, arg_count-2);
   return Item_bool_func::fix_length_and_dec();
 }
@@ -1452,6 +1585,32 @@ null_return:
 }
 
 
+/*
+  This reproduces behavior according to the former
+  Item_func_conv_charset::is_json_type() which returned args[0]->is_json_type().
+  JSON functions with multiple string input with different character sets
+  wrap some arguments into Item_func_conv_charset. So the former
+  Item_func_conv_charset::is_json_type() took the JSON propery from args[0],
+  i.e. from the original argument before the conversion.
+  This is probably not always correct because an *explicit*
+  `CONVERT(arg USING charset)` is actually a general purpose string
+  expression, not a JSON expression.
+*/
+bool is_json_type(const Item *item)
+{
+  for ( ; ; )
+  {
+    if (Type_handler_json_common::is_json_type_handler(item->type_handler()))
+      return true;
+    const Item_func_conv_charset *func;
+    if (!(func= dynamic_cast<const Item_func_conv_charset*>(item)))
+      return false;
+    item= func->arguments()[0];
+  }
+  return false;
+}
+
+
 static int append_json_value(String *str, Item *item, String *tmp_val)
 {
   if (item->type_handler()->is_bool_type())
@@ -1480,20 +1639,20 @@ static int append_json_value(String *str, Item *item, String *tmp_val)
     String *sv= item->val_json(tmp_val);
     if (item->null_value)
       goto append_null;
-    if (item->is_json_type())
+    if (is_json_type(item))
       return str->append(sv->ptr(), sv->length());
 
     if (item->result_type() == STRING_RESULT)
     {
-      return str->append("\"", 1) ||
+      return str->append('"') ||
              st_append_escaped(str, sv) ||
-             str->append("\"", 1);
+             str->append('"');
     }
     return st_append_escaped(str, sv);
   }
 
 append_null:
-  return str->append("null", 4);
+  return str->append(STRING_WITH_LEN("null"));
 }
 
 
@@ -1526,20 +1685,20 @@ static int append_json_value_from_field(String *str,
     String *sv= f->val_str(tmp_val, key + offset);
     if (f->is_null_in_record(key))
       goto append_null;
-    if (i->is_json_type())
+    if (is_json_type(i))
       return str->append(sv->ptr(), sv->length());
 
     if (i->result_type() == STRING_RESULT)
     {
-      return str->append("\"", 1) ||
+      return str->append('"') ||
              st_append_escaped(str, sv) ||
-             str->append("\"", 1);
+             str->append('"');
     }
     return st_append_escaped(str, sv);
   }
 
 append_null:
-  return str->append("null", 4);
+  return str->append(STRING_WITH_LEN("null"));
 }
 
 
@@ -1549,7 +1708,7 @@ static int append_json_keyname(String *str, Item *item, String *tmp_val)
   if (item->null_value)
     goto append_null;
 
-  return str->append("\"", 1) ||
+  return str->append('"') ||
          st_append_escaped(str, sv) ||
          str->append("\": ", 3);
 
@@ -1589,13 +1748,13 @@ bool Item_func_json_array::fix_length_and_dec()
 
 String *Item_func_json_array::val_str(String *str)
 {
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
   uint n_arg;
 
   str->length(0);
   str->set_charset(collation.collation);
 
-  if (str->append("[", 1) ||
+  if (str->append('[') ||
       ((arg_count > 0) && append_json_value(str, args[0], &tmp_val)))
     goto err_return;
 
@@ -1606,7 +1765,7 @@ String *Item_func_json_array::val_str(String *str)
       goto err_return;
   }
 
-  if (str->append("]", 1))
+  if (str->append(']'))
     goto err_return;
 
   if (result_limit == 0)
@@ -1642,7 +1801,7 @@ bool Item_func_json_array_append::fix_length_and_dec()
   }
 
   fix_char_length_ulonglong(char_length);
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -1654,8 +1813,9 @@ String *Item_func_json_array_append::val_str(String *str)
   uint n_arg, n_path;
   size_t str_rest_len;
   const uchar *ar_end;
+  THD *thd= current_thd;
 
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
 
   if ((null_value= args[0]->null_value))
     return 0;
@@ -1681,6 +1841,7 @@ String *Item_func_json_array_append::val_str(String *str)
 
     json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                     (const uchar *) js->ptr() + js->length());
+    je.killed_ptr= (uchar*)&thd->killed;
 
     c_path->cur_step= c_path->p.steps;
 
@@ -1735,11 +1896,11 @@ String *Item_func_json_array_append::val_str(String *str)
       else
         c_to= je.value_end;
 
-      if (str->append("[", 1) ||
+      if (str->append('[') ||
           str->append((const char *) c_from, c_to - c_from) ||
           str->append(", ", 2) ||
           append_json_value(str, args[n_arg+1], &tmp_val) ||
-          str->append("]", 1) ||
+          str->append(']') ||
           str->append((const char *) je.s.c_str,
                       js->end() - (const char *) je.s.c_str))
         goto return_null; /* Out of memory. */
@@ -1761,8 +1922,7 @@ String *Item_func_json_array_append::val_str(String *str)
 
   json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                   (const uchar *) js->ptr() + js->length());
-  str->length(0);
-  str->set_charset(js->charset());
+  je.killed_ptr= (uchar*)&thd->killed;
   if (json_nice(&je, str, Item_func_json_format::LOOSE))
     goto js_error;
 
@@ -1772,6 +1932,7 @@ js_error:
   report_json_error(js, &je, 0);
 
 return_null:
+  thd->check_killed(); // to get the error message right
   null_value= 1;
   return 0;
 }
@@ -1782,8 +1943,9 @@ String *Item_func_json_array_insert::val_str(String *str)
   json_engine_t je;
   String *js= args[0]->val_json(&tmp_js);
   uint n_arg, n_path;
+  THD *thd= current_thd;
 
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
 
   if ((null_value= args[0]->null_value))
     return 0;
@@ -1819,6 +1981,7 @@ String *Item_func_json_array_insert::val_str(String *str)
 
     json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                     (const uchar *) js->ptr() + js->length());
+    je.killed_ptr= (uchar*)&thd->killed;
 
     c_path->cur_step= c_path->p.steps;
 
@@ -1858,7 +2021,7 @@ String *Item_func_json_array_insert::val_str(String *str)
         goto js_error;
     }
 
-    if (unlikely(je.s.error))
+    if (unlikely(je.s.error || *je.killed_ptr))
       goto js_error;
 
     str->length(0);
@@ -1902,8 +2065,7 @@ String *Item_func_json_array_insert::val_str(String *str)
 
   json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                   (const uchar *) js->ptr() + js->length());
-  str->length(0);
-  str->set_charset(js->charset());
+  je.killed_ptr= (uchar*)&thd->killed;
   if (json_nice(&je, str, Item_func_json_format::LOOSE))
     goto js_error;
 
@@ -1912,6 +2074,7 @@ String *Item_func_json_array_insert::val_str(String *str)
 js_error:
   report_json_error(js, &je, 0);
 return_null:
+  thd->check_killed(); // to get the error message right
   null_value= 1;
   return 0;
 }
@@ -1919,13 +2082,13 @@ return_null:
 
 String *Item_func_json_object::val_str(String *str)
 {
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
   uint n_arg;
 
   str->length(0);
   str->set_charset(collation.collation);
 
-  if (str->append("{", 1) ||
+  if (str->append('{') ||
       (arg_count > 0 &&
        (append_json_keyname(str, args[0], &tmp_val) ||
         append_json_value(str, args[1], &tmp_val))))
@@ -1939,7 +2102,7 @@ String *Item_func_json_object::val_str(String *str)
       goto err_return;
   }
 
-  if (str->append("}", 1))
+  if (str->append('}'))
     goto err_return;
 
   if (result_limit == 0)
@@ -1962,6 +2125,16 @@ err_return:
 
 static int do_merge(String *str, json_engine_t *je1, json_engine_t *je2)
 {
+
+  DBUG_EXECUTE_IF("json_check_min_stack_requirement",
+                  {
+                    long arbitrary_var;
+                    long stack_used_up= (available_stack_size(current_thd->thread_stack, &arbitrary_var));
+                    ALLOCATE_MEM_ON_STACK(my_thread_stack_size-stack_used_up-STACK_MIN_SIZE);
+                  });
+  if (check_stack_overrun(current_thd, STACK_MIN_SIZE , NULL))
+    return 1;
+
   if (json_read_value(je1) || json_read_value(je2))
     return 1;
 
@@ -1976,7 +2149,7 @@ static int do_merge(String *str, json_engine_t *je1, json_engine_t *je2)
   
     json_string_set_cs(&key_name, je1->s.cs);
 
-    if (str->append("{", 1))
+    if (str->append('{'))
       return 3;
     while (json_scan_next(je1) == 0 &&
            je1->state != JST_OBJ_END)
@@ -2002,7 +2175,7 @@ static int do_merge(String *str, json_engine_t *je1, json_engine_t *je2)
         *je2= sav_je2;
       }
 
-      if (str->append("\"", 1) ||
+      if (str->append('"') ||
           append_simple(str, key_start, key_end - key_start) ||
           str->append("\":", 2))
         return 3;
@@ -2070,8 +2243,7 @@ merged_j1:
             return 2;
           continue;
         }
-        if (json_skip_key(je2) ||
-            json_skip_level(je1))
+        if (json_skip_key(je2) || json_skip_level(je1))
           return 1;
         goto continue_j2;
       }
@@ -2087,7 +2259,7 @@ merged_j1:
       if (json_skip_key(je2))
         return 1;
 
-      if (str->append("\"", 1) ||
+      if (str->append('"') ||
           append_simple(str, key_start, je2->s.c_str - key_start))
         return 3;
 
@@ -2095,7 +2267,7 @@ continue_j2:
       continue;
     }
 
-    if (str->append("}", 1))
+    if (str->append('}'))
       return 3;
   }
   else
@@ -2115,7 +2287,7 @@ continue_j2:
     }
     else
     {
-      if (str->append("[", 1))
+      if (str->append('['))
         return 3;
       if (je1->value_type == JSON_VALUE_OBJECT)
       {
@@ -2157,7 +2329,7 @@ continue_j2:
       return 3;
 
     if (je2->value_type != JSON_VALUE_ARRAY &&
-        str->append("]", 1))
+        str->append(']'))
       return 3;
   }
 
@@ -2167,10 +2339,11 @@ continue_j2:
 
 String *Item_func_json_merge::val_str(String *str)
 {
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
   json_engine_t je1, je2;
   String *js1= args[0]->val_json(&tmp_js1), *js2=NULL;
   uint n_arg;
+  THD *thd= current_thd;
   LINT_INIT(js2);
 
   if (args[0]->null_value)
@@ -2187,9 +2360,11 @@ String *Item_func_json_merge::val_str(String *str)
 
     json_scan_start(&je1, js1->charset(),(const uchar *) js1->ptr(),
                     (const uchar *) js1->ptr() + js1->length());
+    je1.killed_ptr= (uchar*)&thd->killed;
 
     json_scan_start(&je2, js2->charset(),(const uchar *) js2->ptr(),
                     (const uchar *) js2->ptr() + js2->length());
+    je2.killed_ptr= (uchar*)&thd->killed;
 
     if (do_merge(str, &je1, &je2))
       goto error_return;
@@ -2211,8 +2386,7 @@ String *Item_func_json_merge::val_str(String *str)
 
   json_scan_start(&je1, js1->charset(),(const uchar *) js1->ptr(),
                   (const uchar *) js1->ptr() + js1->length());
-  str->length(0);
-  str->set_charset(js1->charset());
+  je1.killed_ptr= (uchar*)&thd->killed;
   if (json_nice(&je1, str, Item_func_json_format::LOOSE))
     goto error_return;
 
@@ -2224,6 +2398,7 @@ error_return:
     report_json_error(js1, &je1, 0);
   if (je2.s.error)
     report_json_error(js2, &je2, n_arg);
+  thd->check_killed(); // to get the error message right
 null_return:
   null_value= 1;
   return NULL;
@@ -2256,7 +2431,7 @@ static int copy_value_patch(String *str, json_engine_t *je)
   }
   /* JSON_VALUE_OBJECT */
 
-  if (str->append("{", 1))
+  if (str->append('{'))
     return 1;
   while (json_scan_next(je) == 0 && je->state != JST_OBJ_END)
   {
@@ -2279,12 +2454,12 @@ static int copy_value_patch(String *str, json_engine_t *je)
     else
       first_key= 0;
 
-    if (str->append("\"", 1) ||
+    if (str->append('"') ||
         append_simple(str, key_start, je->value_begin - key_start) ||
         copy_value_patch(str, je))
       return 1;
   }
-  if (str->append("}", 1))
+  if (str->append('}'))
     return 1;
 
   return 0;
@@ -2294,6 +2469,15 @@ static int copy_value_patch(String *str, json_engine_t *je)
 static int do_merge_patch(String *str, json_engine_t *je1, json_engine_t *je2,
                           bool *empty_result)
 {
+  DBUG_EXECUTE_IF("json_check_min_stack_requirement",
+                  {
+                    long arbitrary_var;
+                    long stack_used_up= (available_stack_size(current_thd->thread_stack, &arbitrary_var));
+                    ALLOCATE_MEM_ON_STACK(my_thread_stack_size-stack_used_up-STACK_MIN_SIZE);
+                  });
+  if (check_stack_overrun(current_thd, STACK_MIN_SIZE , NULL))
+    return 1;
+
   if (json_read_value(je1) || json_read_value(je2))
     return 1;
 
@@ -2311,7 +2495,7 @@ static int do_merge_patch(String *str, json_engine_t *je1, json_engine_t *je2,
     *empty_result= FALSE;
     json_string_set_cs(&key_name, je1->s.cs);
 
-    if (str->append("{", 1))
+    if (str->append('{'))
       return 3;
     while (json_scan_next(je1) == 0 &&
            je1->state != JST_OBJ_END)
@@ -2337,7 +2521,7 @@ static int do_merge_patch(String *str, json_engine_t *je1, json_engine_t *je2,
         *je2= sav_je2;
       }
 
-      if (str->append("\"", 1) ||
+      if (str->append('"') ||
           append_simple(str, key_start, key_end - key_start) ||
           str->append("\":", 2))
         return 3;
@@ -2428,7 +2612,7 @@ merged_j1:
       if (!first_key && str->append(", ", 2))
         return 3;
 
-      if (str->append("\"", 1) ||
+      if (str->append('"') ||
           append_simple(str, key_start, key_end - key_start) ||
           str->append("\":", 2))
         return 3;
@@ -2449,7 +2633,7 @@ continue_j2:
       continue;
     }
 
-    if (str->append("}", 1))
+    if (str->append('}'))
       return 3;
   }
   else
@@ -2468,11 +2652,12 @@ continue_j2:
 
 String *Item_func_json_merge_patch::val_str(String *str)
 {
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
   json_engine_t je1, je2;
   String *js1= args[0]->val_json(&tmp_js1), *js2=NULL;
   uint n_arg;
   bool empty_result, merge_to_null;
+  THD *thd= current_thd;
 
   /* To report errors properly if some JSON is invalid. */
   je1.s.error= je2.s.error= 0;
@@ -2489,16 +2674,15 @@ String *Item_func_json_merge_patch::val_str(String *str)
 
     json_scan_start(&je2, js2->charset(),(const uchar *) js2->ptr(),
                     (const uchar *) js2->ptr() + js2->length());
+    je2.killed_ptr= (uchar*)&thd->killed;
 
     if (merge_to_null)
     {
       if (json_read_value(&je2))
         goto error_return;
       if (je2.value_type == JSON_VALUE_OBJECT)
-      {
-        merge_to_null= true;
         goto cont_point;
-      }
+
       merge_to_null= false;
       str->set(js2->ptr(), js2->length(), js2->charset());
       goto cont_point;
@@ -2510,12 +2694,13 @@ String *Item_func_json_merge_patch::val_str(String *str)
 
     json_scan_start(&je1, js1->charset(),(const uchar *) js1->ptr(),
                     (const uchar *) js1->ptr() + js1->length());
+    je1.killed_ptr= (uchar*)&thd->killed;
 
     if (do_merge_patch(str, &je1, &je2, &empty_result))
       goto error_return;
 
     if (empty_result)
-      str->append("null");
+      str->append(STRING_WITH_LEN("null"));
 
 cont_point:
     {
@@ -2538,8 +2723,7 @@ cont_point:
 
   json_scan_start(&je1, js1->charset(),(const uchar *) js1->ptr(),
                   (const uchar *) js1->ptr() + js1->length());
-  str->length(0);
-  str->set_charset(js1->charset());
+  je1.killed_ptr= (uchar*)&thd->killed;
   if (json_nice(&je1, str, Item_func_json_format::LOOSE))
     goto error_return;
 
@@ -2551,6 +2735,7 @@ error_return:
     report_json_error(js1, &je1, 0);
   if (je2.s.error)
     report_json_error(js2, &je2, n_arg);
+  thd->check_killed(); // to get the error message right
 null_return:
   null_value= 1;
   return NULL;
@@ -2561,7 +2746,7 @@ bool Item_func_json_length::fix_length_and_dec()
 {
   if (arg_count > 1)
     path.set_constant_flag(args[1]->const_item());
-  maybe_null= 1;
+  set_maybe_null();
   max_length= 10;
   return FALSE;
 }
@@ -2707,7 +2892,7 @@ bool Item_func_json_type::fix_length_and_dec()
 {
   collation.set(&my_charset_utf8mb3_general_ci);
   max_length= 12;
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -2776,7 +2961,7 @@ bool Item_func_json_insert::fix_length_and_dec()
   }
 
   fix_char_length_ulonglong(char_length);
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -2787,8 +2972,9 @@ String *Item_func_json_insert::val_str(String *str)
   String *js= args[0]->val_json(&tmp_js);
   uint n_arg, n_path;
   json_string_t key_name;
+  THD *thd= current_thd;
 
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
 
   if ((null_value= args[0]->null_value))
     return 0;
@@ -2827,6 +3013,7 @@ String *Item_func_json_insert::val_str(String *str)
 
     json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                     (const uchar *) js->ptr() + js->length());
+    je.killed_ptr= (uchar*)&thd->killed;
 
     if (c_path->p.last_step < c_path->p.steps)
       goto v_found;
@@ -2876,7 +3063,7 @@ String *Item_func_json_insert::val_str(String *str)
         str->length(0);
         /* Wrap the value as an array. */
         if (append_simple(str, js->ptr(), (const char *) v_from - js->ptr()) ||
-            (do_array_autowrap && str->append("[", 1)))
+            (do_array_autowrap && str->append('[')))
           goto js_error; /* Out of memory. */
 
         if (je.value_type == JSON_VALUE_OBJECT)
@@ -2889,7 +3076,7 @@ String *Item_func_json_insert::val_str(String *str)
              (append_simple(str, v_from, je.s.c_str - v_from) ||
               str->append(", ", 2))) ||
             append_json_value(str, args[n_arg+1], &tmp_val) ||
-            (do_array_autowrap && str->append("]", 1)) ||
+            (do_array_autowrap && str->append(']')) ||
             append_simple(str, je.s.c_str, js->end()-(const char *) je.s.c_str))
           goto js_error; /* Out of memory. */
 
@@ -2960,7 +3147,7 @@ String *Item_func_json_insert::val_str(String *str)
       str->length(0);
       if (append_simple(str, js->ptr(), v_to - js->ptr()) ||
           (n_key > 0 && str->append(", ", 2)) ||
-          str->append("\"", 1) ||
+          str->append('"') ||
           append_simple(str, lp->key, lp->key_end - lp->key) ||
           str->append("\":", 2) ||
           append_json_value(str, args[n_arg+1], &tmp_val) ||
@@ -3008,7 +3195,7 @@ continue_point:
 
   json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                   (const uchar *) js->ptr() + js->length());
-  str->length(0);
+  je.killed_ptr= (uchar*)&thd->killed;
   if (json_nice(&je, str, Item_func_json_format::LOOSE))
     goto js_error;
 
@@ -3016,6 +3203,7 @@ continue_point:
 
 js_error:
   report_json_error(js, &je, 0);
+  thd->check_killed(); // to get the error message right
 return_null:
   null_value= 1;
   return 0;
@@ -3028,7 +3216,7 @@ bool Item_func_json_remove::fix_length_and_dec()
   max_length= args[0]->max_length;
 
   mark_constant_paths(paths, args+1, arg_count-1);
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -3039,8 +3227,9 @@ String *Item_func_json_remove::val_str(String *str)
   String *js= args[0]->val_json(&tmp_js);
   uint n_arg, n_path;
   json_string_t key_name;
+  THD *thd= current_thd;
 
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
 
   if (args[0]->null_value)
     goto null_return;
@@ -3085,6 +3274,7 @@ String *Item_func_json_remove::val_str(String *str)
 
     json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                     (const uchar *) js->ptr() + js->length());
+    je.killed_ptr= (uchar*)&thd->killed;
 
     c_path->cur_step= c_path->p.steps;
 
@@ -3193,8 +3383,7 @@ v_found:
 
   json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                   (const uchar *) js->ptr() + js->length());
-  str->length(0);
-  str->set_charset(js->charset());
+  je.killed_ptr= (uchar*)&thd->killed;
   if (json_nice(&je, str, Item_func_json_format::LOOSE))
     goto js_error;
 
@@ -3202,6 +3391,7 @@ v_found:
   return str;
 
 js_error:
+  thd->check_killed(); // to get the error message right
   report_json_error(js, &je, 0);
 null_return:
   null_value= 1;
@@ -3213,7 +3403,7 @@ bool Item_func_json_keys::fix_length_and_dec()
 {
   collation.set(args[0]->collation);
   max_length= args[0]->max_length;
-  maybe_null= 1;
+  set_maybe_null();
   if (arg_count > 1)
     path.set_constant_flag(args[1]->const_item());
   return FALSE;
@@ -3305,7 +3495,7 @@ skip_search:
     goto null_return;
   
   str->length(0);
-  if (str->append("[", 1))
+  if (str->append('['))
     goto err_return; /* Out of memory. */
   /* Parse the OBJECT collecting the keys. */
   while (json_scan_next(&je) == 0 && je.state != JST_OBJ_END)
@@ -3328,9 +3518,9 @@ skip_search:
       if (!check_key_in_list(str, key_start, key_len))
       { 
         if ((n_keys > 0 && str->append(", ", 2)) ||
-          str->append("\"", 1) ||
+          str->append('"') ||
           append_simple(str, key_start, key_len) ||
-          str->append("\"", 1))
+          str->append('"'))
         goto err_return;
         n_keys++;
       }
@@ -3345,7 +3535,7 @@ skip_search:
     }
   }
 
-  if (unlikely(je.s.error || str->append("]", 1)))
+  if (unlikely(je.s.error || str->append(']')))
     goto err_return;
 
   null_value= 0;
@@ -3398,7 +3588,7 @@ bool Item_func_json_search::fix_length_and_dec()
 
   if (arg_count > 4)
     mark_constant_paths(paths, args+4, arg_count-4);
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -3449,14 +3639,14 @@ static int append_json_path(String *str, const json_path_t *p)
     else /*JSON_PATH_ARRAY*/
     {
 
-      if (str->append("[", 1) ||
+      if (str->append('[') ||
           str->append_ulonglong(c->n_item) ||
-          str->append("]", 1))
+          str->append(']'))
         return TRUE;
     }
   }
 
-  return str->append("\"", 1);
+  return str->append('"');
 }
 
 
@@ -3517,7 +3707,7 @@ String *Item_func_json_search::val_str(String *str)
         {
           if (n_path_found == 2)
           {
-            if (str->append("[", 1) ||
+            if (str->append('[') ||
                 append_json_path(str, &sav_path))
                 goto js_error;
           }
@@ -3543,7 +3733,7 @@ end:
   }
   else
   {
-    if (str->append("]", 1))
+    if (str->append(']'))
       goto js_error;
   }
 
@@ -3559,29 +3749,30 @@ null_return:
 }
 
 
-const char *Item_func_json_format::func_name() const
+LEX_CSTRING Item_func_json_format::func_name_cstring() const
 {
   switch (fmt)
   {
   case COMPACT:
-    return "json_compact";
+    return { STRING_WITH_LEN("json_compact") };
   case LOOSE:
-    return "json_loose";
+    return { STRING_WITH_LEN("json_loose") };
   case DETAILED:
-    return "json_detailed";
+    return { STRING_WITH_LEN("json_detailed") };
   default:
     DBUG_ASSERT(0);
   };
 
-  return "";
+  return NULL_clex_str;
 }
 
 
 bool Item_func_json_format::fix_length_and_dec()
 {
   decimals= 0;
+  collation.set(args[0]->collation);
   max_length= args[0]->max_length;
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -3591,6 +3782,7 @@ String *Item_func_json_format::val_str(String *str)
   String *js= args[0]->val_json(&tmp_js);
   json_engine_t je;
   int tab_size= 4;
+  THD *thd= current_thd;
 
   if ((null_value= args[0]->null_value))
     return 0;
@@ -3614,13 +3806,13 @@ String *Item_func_json_format::val_str(String *str)
 
   json_scan_start(&je, js->charset(), (const uchar *) js->ptr(),
                   (const uchar *) js->ptr()+js->length());
+  je.killed_ptr= (uchar*)&thd->killed;
 
-  str->length(0);
-  str->set_charset(js->charset());
   if (json_nice(&je, str, fmt, tab_size))
   {
     null_value= 1;
     report_json_error(js, &je, 0);
+    thd->check_killed(); // to get the error message right
     return 0;
   }
 
@@ -3754,6 +3946,12 @@ void Item_func_json_arrayagg::cut_max_length(String *result,
 }
 
 
+Item *Item_func_json_arrayagg::copy_or_same(THD* thd)
+{
+   return new (thd->mem_root) Item_func_json_arrayagg(thd, this);
+}
+
+
 String* Item_func_json_arrayagg::val_str(String *str)
 {
   if ((str= Item_func_group_concat::val_str(str)))
@@ -3774,7 +3972,7 @@ Item_func_json_objectagg(THD *thd, Item_func_json_objectagg *item)
 {
   quick_group= FALSE;
   result.set_charset(collation.collation);
-  result.append("{");
+  result.append('{');
 }
 
 
@@ -3782,14 +3980,14 @@ bool
 Item_func_json_objectagg::fix_fields(THD *thd, Item **ref)
 {
   uint i;                       /* for loop variable */
-  DBUG_ASSERT(fixed == 0);
+  DBUG_ASSERT(fixed() == 0);
 
   memcpy(orig_args, args, sizeof(Item*) * arg_count);
 
   if (init_sum_func_check(thd))
     return TRUE;
 
-  maybe_null= 1;
+  set_maybe_null();
 
   /*
     Fix fields for select list and ORDER clause
@@ -3799,9 +3997,7 @@ Item_func_json_objectagg::fix_fields(THD *thd, Item **ref)
   {
     if (args[i]->fix_fields_if_needed_for_scalar(thd, &args[i]))
       return TRUE;
-    m_with_subquery|= args[i]->with_subquery();
-    with_param|= args[i]->with_param;
-    with_window_func|= args[i]->with_window_func;
+    with_flags|= args[i]->with_flags;
   }
 
   /* skip charset aggregation for order columns */
@@ -3818,7 +4014,7 @@ Item_func_json_objectagg::fix_fields(THD *thd, Item **ref)
   if (check_sum_func(thd, ref))
     return TRUE;
 
-  fixed= 1;
+  base_flags|= item_base_t::FIXED;
   return FALSE;
 }
 
@@ -3857,11 +4053,11 @@ bool Item_func_json_objectagg::add()
 
   null_value= 0;
   if (result.length() > 1)
-    result.append(", ");
+    result.append(STRING_WITH_LEN(", "));
 
-  result.append("\"");
+  result.append('"');
   result.append(*key);
-  result.append("\":");
+  result.append(STRING_WITH_LEN("\":"));
 
   buf.length(0);
   append_json_value(&result, args[1], &buf);
@@ -3872,11 +4068,11 @@ bool Item_func_json_objectagg::add()
 
 String* Item_func_json_objectagg::val_str(String* str)
 {
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
   if (null_value)
     return 0;
 
-  result.append("}");
+  result.append('}');
   return &result;
 }
 
